"""
A quest is a distributed scraping journey :)
A sciop instance contains a list of dataset parts to scrape,
a crawler will get a series of unclaimed parts,
crawl them, and upload the resulting torrents!
"""

import json
import warnings
from pathlib import Path
from typing import Any, ClassVar, Literal, Self

from pydantic import BaseModel, Field
from scrapy import Spider
from scrapy.crawler import CrawlerProcess


class Quest(BaseModel):
    """
    A scrape journey to go on with your friends
    """

    name: ClassVar[str]
    """
    Name that will be used when calling from the CLI.
    Each quest needs an (instance) unique name
    """
    dataset_slug: ClassVar[str]
    """
    Dataset slug on the sciop instance that this quest belongs to.
    
    Usually the same as `name`, but allowed to differ e.g.
    in the case of multiple quests for the same dataset.
    """
    spider: ClassVar[type[Spider] | None] = None
    """
    Scrapy spider to use, if any
    """

    output: Path
    """
    Output directory to write to.
    
    The output directory stores data for *all* quests and subquests.
    
    Quests should ensure that their scraper outputs data in a structure like..
    
    ```
    output
    |- quest_status.json
    |- quest_1
    |  |- subquest_1
    |  |  |- (data)
    |  |  ...
    |  |- subquest_1
    |  ...
    |- quest_2
    ...
    ```
    """
    subquest: str
    """
    The currently active subquest.
    """

    def run(self, scrape_kwargs: dict | None = None) -> "QuestStatus":
        if scrape_kwargs is None:
            scrape_kwargs = {}

        status = self.init_status()
        try:
            if status.status == "scraping":
                status, scrape_kwargs = self.before_scrape(status, **scrape_kwargs)
                status = self.scrape(status, **scrape_kwargs)
                status = self.after_scrape(status)
                status.status = "validating"
            if status.status == "validating":
                status = self.before_validate(status)
                status = self.validate_scrape(status)
                status = self.after_validate(status)
                status.status = "packing"
            if status.status == "packing":
                status = self.before_pack(status)
                status = self.pack(status)
                status = self.after_pack(status)
                status.status = "complete"
            if status.result is None:
                # status would otherwise be set to validation_error in validate_scrape
                status.result = "success"
        finally:
            self.update_log(status)

        return status

    def init_status(self) -> "QuestStatus":
        """
        Load any intermediate result from the quest log,
        or create a new one if none found.

        If we have previously attempted a subquest and ended with a validation error,
        retry from the scraping stage.
        """
        status = None
        if self.log_path.exists():
            log = QuestLog.from_json(self.log_path)
            status = log.get_subquest(self.name, self.subquest)

        if status is None:
            status = QuestStatus(
                quest=self.name,
                subquest=self.subquest,
                path=self.subquest_path,
            )
        else:
            # if we have previously completed with a validation error and are running again,
            # return to scraping state.
            if status.status == "complete" and status.result == "validation_error":
                status.status = "scraping"

        return status

    @property
    def log_path(self) -> Path:
        return self.output / "quest-log.json"

    @property
    def subquest_path(self) -> Path:
        return self.output / self.name / self.subquest

    def update_log(self, res: "QuestStatus") -> None:
        """
        Update the log, replacing the
        :param res:
        :return:
        """
        log = QuestLog.from_json(self.log_path)
        log = log.update(res)
        log.to_json(self.log_path)

    def scrape(self, res: "QuestStatus", **kwargs: Any) -> "QuestStatus":
        """
        Run the spider - default is to use the scrapy spider
        """
        return self._scrape_with_spider(res, **kwargs)

    def validate_scrape(self, res: "QuestStatus") -> "QuestStatus":
        """
        After scraping, subclasses may override to perform validation.

        Default is no-op
        """
        return res

    def pack(self, res: "QuestStatus") -> "QuestStatus":
        """
        After validating, subclasses may override to perform packing.

        Default is no-op
        """
        return res

    def before_scrape(self, res: "QuestStatus", **kwargs: any) -> tuple["QuestStatus", dict]:
        """Hook method called before scraping"""
        return res, kwargs

    def after_scrape(self, res: "QuestStatus") -> "QuestStatus":
        """Hook method called after scraping"""
        return res

    def before_validate(self, res: "QuestStatus") -> "QuestStatus":
        """Hook method called before validation"""
        return res

    def after_validate(self, res: "QuestStatus") -> "QuestStatus":
        """Hook method called after validation"""
        return res

    def before_pack(self, res: "QuestStatus") -> "QuestStatus":
        """Hook method called before packing"""
        return res

    def after_pack(self, res: "QuestStatus") -> "QuestStatus":
        """Hook method called after packing"""
        return res

    def _scrape_with_spider(self, res: "QuestStatus", **kwargs: Any) -> None:
        if self.spider is None:
            raise RuntimeError("No spider has been declared, but attempted to scrape with spider")

        process = CrawlerProcess()
        process.crawl(self.spider, **kwargs)
        process.start()
        return res


class QuestStatus(BaseModel):
    quest: str
    subquest: str
    path: Path
    """Directory where the data was scraped to"""
    scrape_errors: list["ScrapeError"] = Field(default_factory=list)
    validation_errors: list["ValidationError"] = Field(default_factory=list)
    status: Literal["scraping", "validating", "packing", "complete"] = "scraping"
    result: None | Literal["success", "validation_error"] = None


class QuestLog(BaseModel):
    subquests: list[QuestStatus] = Field(default_factory=list)

    def get_subquest(self, quest: str, subquest: str) -> QuestStatus | None:
        matches = [q for q in self.subquests if q.quest == quest and q.subquest == subquest]
        if len(matches) > 1:
            raise KeyError(f"More than one match found for quest {quest}, subquest {subquest}")
        elif len(matches) == 1:
            return matches[0]
        else:
            return None

    def update(self, res: "QuestStatus") -> Self:
        """
        Add new entry to log, replacing any previous matches
        """
        items = [q for q in self.subquests if q.quest != res.quest and q.subquest != res.subquest]
        items.append(res)
        self.subquests = items
        return self

    @classmethod
    def from_json(cls, path: Path) -> "QuestLog":
        if not path.exists():
            return QuestLog()

        try:
            with open(path) as f:
                log = json.load(f)
        except json.decoder.JSONDecodeError:
            warnings.warn(
                f"Quest log could not be read from {str(path)}, ignoring, will overwrite",
                stacklevel=2,
            )
            log = []
        return QuestLog(subquests=log)

    def to_json(self, path: Path) -> None:
        items = self.model_dump(by_alias=True)["subquests"]
        with open(path, "w") as f:
            json.dump(items, f, indent=2, default=str)


class ScrapeError(BaseModel):
    url: str
    type_: str = Field(..., alias="type")
    msg: str


class ValidationError(BaseModel):
    path: Path
    type_: str = Field(..., alias="type")
    msg: str



from pathlib import Path

import click

from sciop_scraping.quests.chronicling import ChroniclingAmericaQuest


@click.command("chronicling-america")
@click.option("-b", "--batch", help="Which batch to crawl. If None, crawl everything", default=None)
@click.option(
    "-o",
    "--output",
    help="Output directory to save files in. "
    "If None, $PWD/data/chronicling-america. "
    "Data will be saved in a chronicling-america subdirectory, "
    "and the crawl state will be saved in crawl_state.",
    default=None,
    type=click.Path(),
)
@click.option(
    "-c",
    "--cloudflare-cookie",
    help="When you get rate limited, you need to go solve a cloudflare challenge, "
    "grab the cookie with the key 'cf_clearance' and pass it here",
    default=None,
)
@click.option(
    "-u",
    "--user-agent",
    help="When you get rate limited, the cookie is tied to a specific user agent, "
    "copy paste that and pass it here",
    default=None,
)
@click.option(
    "--crawl-state",
    help="Use scrapy crawl state. Defaults False, "
    "because we can resume crawling using the manifest.",
    default=False,
    is_flag=True,
)
def chronicling_america(
    batch: str | None,
    output: Path | None = None,
    cloudflare_cookie: str | None = None,
    user_agent: str | None = None,
    crawl_state: bool = False,
) -> None:
    """
    Scrape the Chronicling America dataset from the Library of Congress in batches

    https://chroniclingamerica.loc.gov/data/batches/

    If you get a 429 redirect, you will need to manually bypass the cloudflare ratelimit check.

    - Open https://chroniclingamerica.loc.gov/data/batches/ in a browser,
    - Pass the cloudflare check
    - Open your developer tools (often right click + inspect element)
    - Open the networking tab to watch network requests
    - Reload the page
    - Click on the request made to the page you're on to see the request headers
    - Copy your user agent and the part of the cookie after `cf_clearance=`
      and pass them to the -u and -c cli options, respectively.
    """
    if output is None:
        output = Path.cwd() / "data"

    job_dir = None
    if crawl_state:
        job_dir = output / "crawl_state" / batch
        job_dir.mkdir(exist_ok=True, parents=True)

    quest = ChroniclingAmericaQuest(subquest=batch, output=output)
    quest.run(
        scrape_kwargs={
            "job_dir": job_dir,
            "user_agent": user_agent,
            "batch": batch,
            "cf_cookie": cloudflare_cookie,
        }
    )



import random
import re
from collections.abc import AsyncIterator, Generator
from pathlib import Path
from typing import Any
from urllib.parse import urljoin

import scrapy
from scrapy.http import HtmlResponse, Response, TextResponse
from scrapy.settings import BaseSettings
from scrapy.spidermiddlewares.httperror import HttpError
from twisted.python.failure import Failure

from sciop_scraping.const import USER_AGENTS

BASE_URL = "https://chroniclingamerica.loc.gov/data/batches/"


class ChroniclingAmericaSpider(scrapy.Spider):
    """
    Crawl chronicling america data from the Library of Congress,
    either by batch or as a whole.

    Excludes `.tif` and `.tiff` files, since they are redundant with the .jp2 files.
    """

    name = "chronicling-america"
    allowed_domains = ["chroniclingamerica.loc.gov", "tile.loc.gov"]
    BAGIT_FILES = ["bag-info.txt", "bagit.txt", "manifest-md5.txt", "tagmanifest-md5.txt"]

    USER_AGENT_OVERRIDE: str | None = None
    JOBDIR_OVERRIDE: str | None = None

    def __init__(
        self,
        batch: str | None = None,
        output: Path | None = None,
        cf_cookie: str | None = None,
        download_timeout: int = 10,
        *args: Any,
        **kwargs: Any,
    ):
        super().__init__(*args, **kwargs)
        self.batch = batch
        if output is None:
            self.output = Path.cwd() / "data" / "chronicling-america"
        else:
            self.output = Path(output).resolve()
        self.output.mkdir(exist_ok=True, parents=True)
        self.cf_cookie = {"cf_clearance": cf_cookie} if cf_cookie else None
        if not self.batch:
            self.crawl_root = BASE_URL
            self.start_urls = [BASE_URL]
        else:
            self.crawl_root = BASE_URL + self.batch + "/"
            self.start_urls = [urljoin(self.crawl_root, f) for f in self.BAGIT_FILES]
        self.download_timeout = download_timeout

    async def start(self) -> AsyncIterator[Any]:

        for url in self.start_urls:
            if url.split("/")[-1] == "manifest-md5.txt":
                yield scrapy.Request(
                    self.tile_url(url),
                    callback=self.parse_manifest,
                    dont_filter=True,
                    cookies=self.cf_cookie,
                    errback=self.errback,
                    cb_kwargs={"batch_url": self.crawl_root},
                )
            elif url.split("/")[-1] in self.BAGIT_FILES:
                yield scrapy.Request(
                    self.tile_url(url),
                    callback=self.parse_file,
                    dont_filter=True,
                    cookies=self.cf_cookie,
                    errback=self.errback,
                    cb_kwargs={"original_url": url},
                )
            else:
                yield scrapy.Request(
                    url, dont_filter=True, cookies=self.cf_cookie, errback=self.errback
                )

    @classmethod
    def update_settings(cls, settings: BaseSettings) -> None:
        super().update_settings(settings)
        if cls.USER_AGENT_OVERRIDE:
            settings.set("USER_AGENT", cls.USER_AGENT_OVERRIDE, priority="spider")
        else:
            settings.set("USER_AGENT", random.choice(USER_AGENTS), priority="spider")
        if cls.JOBDIR_OVERRIDE:
            settings.set("JOBDIR", str(cls.JOBDIR_OVERRIDE))
        settings.set("ROBOTSTXT_OBEY", False, priority="spider")
        # there doesn't appear to be a rate limit on tile.loc.gov,
        # and since we can jump straight there when we have a batch name,
        # and i'm not sure how to dynamically modify this, leave this off by default.
        # settings.set("DOWNLOAD_DELAY", 0.5, priority="spider")
        settings.set("RETRY_ENABLED", True, priority="spider")
        settings.set("RETRY_TIMES", 5, priority="spider")
        settings.set("RETRY_HTTP_CODES", [500, 502, 503, 504, 522, 524, 408])

    def url_to_path(self, url: str) -> Path:
        """Get the output path for a given URL."""
        out_name = re.sub(BASE_URL, "", url)
        out_path = self.output / out_name
        out_path.parent.mkdir(exist_ok=True, parents=True)
        return out_path

    def tile_url(self, url: str) -> str:
        """Convert a chroniclingamerica.loc.gov url to a tile.loc.gov url"""
        if "tile.loc.gov" in url:
            return url
        pattern = re.compile(
            r"^https://chroniclingamerica\.loc\.gov/data/batches/(?P<batch>\w+)/(?P<path>.*)"
        )
        match = pattern.match(url)
        if not match:
            raise ValueError(f"Could not convert url to tile url: {url}")
        val = match.groupdict()
        prefix = val["batch"].split("_")[0]
        return f"https://tile.loc.gov/storage-services/service/ndnp/{prefix}/batch_{val['batch']}/{val['path']}"

    def parse(self, response: HtmlResponse) -> Generator[scrapy.Request, None, None]:
        links = response.css("a::attr(href)").getall()

        # if we are on a page that has a `manifest-md5.txt` in it,
        # shortcut and just use that to enumerate the files
        if any([link == "manifest-md5.txt" for link in links]):
            link = urljoin(response.url, "manifest-md5.txt")
            yield response.follow(
                link,
                callback=self.parse_manifest,
                errback=self.errback,
                cookies=self.cf_cookie,
                cb_kwargs={"batch_url": response.url},
            )
        else:
            links = [urljoin(response.url, link) for link in links]
            # filter to only those underneath the crawl root
            links = [link for link in links if self.crawl_root in link]

            for link in links:
                if link.endswith("/"):
                    yield response.follow(
                        link, callback=self.parse, errback=self.errback, cookies=self.cf_cookie
                    )
                else:
                    if not self.url_to_path(link).exists():
                        tile_link = self.tile_url(link)
                        yield response.follow(
                            tile_link,
                            callback=self.parse_file,
                            errback=self.errback,
                            cb_kwargs={"original_url": link},
                            cookies=self.cf_cookie,
                        )
                    else:
                        self.logger.info("Skipping %s, already downloaded", link)

    def parse_manifest(self, response: TextResponse, batch_url: str) -> None:
        lines = response.text.split("\n")
        paths = [re.split(r"\s+", line)[-1] for line in lines]
        # exclude .tif and .tiff files
        paths = [p for p in paths if not p.endswith(".tif") and not p.endswith(".tiff")]
        urls = [urljoin(batch_url, path) for path in paths]
        # yeah this is a little inefficient but we only need to do it once
        existing = [url for url in urls if self.url_to_path(url).exists()]
        not_existing = [url for url in urls if not self.url_to_path(url).exists()]
        if existing:
            self.logger.info(f"Skipping {len(existing)} files, already downloaded")
        self.logger.info(f"Downloading {len(not_existing)} files")
        self.parse_file(response, original_url=urljoin(batch_url, "manifest-md5.txt"))
        for url in not_existing:
            tile_url = self.tile_url(url)
            yield response.follow(
                tile_url,
                callback=self.parse_file,
                errback=self.errback,
                cb_kwargs={"original_url": url},
                cookies=self.cf_cookie,
            )

    def parse_file(self, response: Response, original_url: str) -> None:
        """
        Save a file underneath ``self.output``, removing the base url
        """

        out_path = self.url_to_path(original_url)
        self.logger.debug("Saving %s to %s", response.url, out_path)
        with open(out_path, "wb") as f:
            f.write(response.body)

    def errback(self, failure: Failure) -> None:
        self.logger.error(f"A spider error occurred: {failure}")
        if failure.check(HttpError) and failure.value.response.status == 429:
            raise scrapy.exceptions.CloseSpider(
                "429 received - stopping scrape. We can't wait these out. \n"
                "You should \n"
                "- Open one of the recent pages in a browser,\n"
                "- Pass the cloudflare check\n"
                "- Open your developer tools (often right click + inspect element)\n"
                "- Open the networking tab to watch network requests\n"
                "- Reload the page\n"
                "- Click on the request made to the page you're on to see the request headers\n"
                "- Copy your user agent and the part of the cookie after `cf_clearance=` "
                "and pass them to the -u and -c cli options, respectively.\n"
                "If you are crawling *all* data, rather than a single batch, "
                "you will likely need to set DOWNLOAD_DELAY=1 until the manifests are scraped."
            )



"""
Helper methods for validation that might be used in multiple quests
"""

import hashlib
import re
from pathlib import Path

from tqdm import tqdm

from sciop_scraping.quests.base import ValidationError


def validate_bagit_manifest(path: Path, hash_type: str = "md5") -> list[ValidationError]:
    """
    Given the base directory of a bagit directory that contains `manifest-{hash_type}.txt`
    and `data/` check the files against the manifest,
    returning ValidationErrors for missing or incorrect files.
    """
    errors = []
    manifest_path = path / f"manifest-{hash_type}.txt"

    if not manifest_path.exists():
        errors.append(
            ValidationError(
                type="manifest",
                path=manifest_path,
                msg="No manifest file found at expected location!",
            )
        )
        return errors

    with open(manifest_path) as f:
        manifest = f.read()

    lines = manifest.splitlines()
    # split into (hash, path) pairs
    items = [re.split(r"\s+", line.strip()) for line in lines]

    for item in tqdm(items, desc="Validating files"):
        expected_hash, sub_path = item
        abs_path = path / sub_path
        if not abs_path.exists():
            errors.append(ValidationError(type="missing", path=sub_path, msg="File not found"))
            continue

        with open(abs_path, "rb") as f:
            file_hash = hashlib.file_digest(f, hash_type).hexdigest()

        if file_hash != expected_hash:
            errors.append(ValidationError(type="incorrect", path=sub_path, msg="Hash mismatch"))
    return errors



import multiprocessing

BOT_NAME = "sciop_scrape"

SPIDER_MODULES = ["sciop_scraping.spiders"]
NEWSPIDER_MODULE = "rescue.spiders"

CONCURRENT_REQUESTS = multiprocessing.cpu_count() * 2

RETRY_ENABLED = True


# Set settings whose default value is deprecated to a future-proof value
TWISTED_REACTOR = "twisted.internet.asyncioreactor.AsyncioSelectorReactor"
FEED_EXPORT_ENCODING = "utf-8"



from typing import Any

import scrapy
from scrapy import Request
from scrapy.downloadermiddlewares.retry import RetryMiddleware
from scrapy.http import Response
from scrapy.utils.response import response_status_message
from twisted.internet import defer, reactor


async def async_sleep(delay: float, return_value: Any | None = None) -> defer.Deferred:
    deferred = defer.Deferred()
    reactor.callLater(delay, deferred.callback, return_value)
    return await deferred


class TooManyRequestsRetryMiddleware(RetryMiddleware):
    """
    Modifies RetryMiddleware to delay retries on status 429.

    References:
        https://stackoverflow.com/a/66131114/13113166
    """

    DEFAULT_DELAY = 60  # Delay in seconds.
    MAX_DELAY = 600  # Sometimes, RETRY-AFTER has absurd values

    async def process_response(
        self, request: scrapy.Request, response: Response, spider: scrapy.Spider
    ) -> Response | Request | None:
        """
        Like RetryMiddleware.process_response, but, if response status is 429,
        retry the request only after waiting at most self.MAX_DELAY seconds.
        Respect the Retry-After header if it's less than self.MAX_DELAY.
        If Retry-After is absent/invalid, wait only self.DEFAULT_DELAY seconds.
        """

        if request.meta.get("dont_retry", False):
            return response

        if response.status in self.retry_http_codes:
            if response.status == 429:
                retry_after = response.headers.get("retry-after")
                try:
                    retry_after = int(retry_after)
                except (ValueError, TypeError):
                    delay = self.DEFAULT_DELAY
                else:
                    delay = min(self.MAX_DELAY, retry_after)
                spider.logger.info(f"Retrying {request} in {delay} seconds.")

                spider.crawler.engine.pause()
                await async_sleep(delay)
                spider.crawler.engine.unpause()

            reason = response_status_message(response.status)
            return self._retry(request, reason, spider) or response

        return response



from pathlib import Path
from typing import Any

from sciop_scraping.quests import Quest, QuestStatus
from sciop_scraping.quests.chronicling.spider import ChroniclingAmericaSpider
from sciop_scraping.validation import validate_bagit_manifest


class ChroniclingAmericaQuest(Quest):
    name = "chronicling-america"
    dataset_slug = "chronicling-america"
    spider = ChroniclingAmericaSpider

    @property
    def subquest_path(self) -> Path:
        return self.output / self.name / self.subquest.replace("-", "_")

    def before_scrape(self, res: "QuestStatus", **kwargs: Any) -> tuple["QuestStatus", dict]:
        """Set classvars on spider before scraping, add output dir to kwargs"""

        self.spider.USER_AGENT_OVERRIDE = kwargs.get("user_agent")
        self.spider.JOBDIR_OVERRIDE = kwargs.get("job_dir")

        # URLS contain the batch/subquest name, so we give the parent dir to the spider
        kwargs["output"] = self.subquest_path.parent
        return res, kwargs

    def validate_scrape(self, res: QuestStatus) -> QuestStatus:
        """Compare the crawled files to the md5-manifest"""
        errors = validate_bagit_manifest(self.subquest_path, "md5")

        # since we exclude .tif files from the scrape, validation errors don't count here
        errors = [e for e in errors if e.path.suffix not in (".tif", ".tiff")]

        res.validation_errors = errors
        if errors:
            res.result = "validation_error"
        else:
            res.result = "success"
        return res
